use defmt::*;
use embassy_futures::select::*;
use embassy_futures::yield_now;
use embassy_rp::gpio;
use embassy_rp::pac;
use embassy_rp::interrupt::UART0_IRQ;
use embassy_rp::uart::BufferedUartRx;
use embassy_rp::uart::Parity;
use embassy_time::{Duration, Timer};
use embassy_embedded_hal::SetConfig;
use embassy_rp::peripherals::{self, UART0};
use embassy_rp::uart::{self};
use embassy_rp::pio::{Config, Pio, ShiftConfig, ShiftDirection, FifoJoin};
use embassy_rp::relocate::RelocatedProgram;
use embassy_rp::Peripheral;
use embedded_io::asynch::BufRead;
use {defmt_rtt as _, panic_probe as _};
use embedded_io::asynch::Read;
use fixed::traits::ToFixed;
use fixed::types::U24F8;
use fixed::types::U56F8;

use crate::dont_abort::DontAbort;
use crate::dont_abort::DontAbortMode::*;

pub trait RS485Handler {
    //type CommandFuture: Future<Output = ()>;
    const TX_BUF_LENGTH: usize;

    fn on_rx_frame<F>(self: &mut Self, rx: Result<&[u8], uart::Error>, reply: Option<F>)
        where F: FnOnce(&[u8]);
    fn on_tx_done(self: &mut Self);
    fn on_autobaud_success(self: &mut Self, baudrate: f32);
}

pub const RX_BUF_LEN: usize = 256;

pub struct RS485<H: RS485Handler> {
    pio: peripherals::PIO0,
    rx_pin: peripherals::PIN_17,
    tx_pin: peripherals::PIN_16,
    tx_en_pin: peripherals::PIN_15,
    dma_channel: peripherals::DMA_CH0,
    tx_dma_channel: peripherals::DMA_CH2,
    rx: BufferedUartRx<'static, UART0>,
    rx_buffer: [u8; RX_BUF_LEN],
    uart_baud_rate: u32,
    uart_parity: Parity,
    handler: H
}

fn pin_io<P: gpio::Pin>(pin: &P) -> pac::io::Gpio {
    use gpio::Bank;

    let block = match pin.bank() {
        Bank::Bank0 => pac::IO_BANK0,
        Bank::Qspi => pac::IO_QSPI,
    };
    block.gpio(pin.pin() as _)
}

fn pin_pad<P: gpio::Pin>(pin: &P) -> pac::common::Reg<pac::pads::regs::GpioCtrl, pac::common::RW> {
    use gpio::Bank;

    let block = match pin.bank() {
        Bank::Bank0 => pac::PADS_BANK0,
        Bank::Qspi => pac::PADS_QSPI,
    };
    block.gpio(pin.pin() as _)
}

#[allow(dead_code)]
async fn debug_print_pio_addr(sm: pac::pio::StateMachine) {
    let mut prev = 42u8;
    loop {
        let addr = unsafe { sm.addr().read().addr() };
        if prev != addr {
            prev = addr;
            info!("SM addr: {}", addr);
        }
        Timer::after(Duration::from_millis(200)).await;
    }
}

impl<H: RS485Handler> RS485<H> {
    pub fn new(
            uart: UART0, rx_pin: peripherals::PIN_17, tx_pin: peripherals::PIN_16, tx_en_pin: peripherals::PIN_15,
            rx_irq: UART0_IRQ, _rx_dma_channel: peripherals::DMA_CH1, uart_config: uart::Config, rx_buffer: &'static mut [u8; RX_BUF_LEN],
            pio: peripherals::PIO0, dma_channel: peripherals::DMA_CH0, tx_dma_channel: peripherals::DMA_CH2,
            handler: H
        ) -> RS485<H> {
        // SAFETY: The auto-baud will only read from this pin and we will set it back to UART mode after initialising the PIO.
        let rx_pin_for_autobaud = unsafe { rx_pin.clone_unchecked() };

        core::assert!(rx_buffer.len() == RX_BUF_LEN,
            "The buffers must have the same length because BufferedUartRx::try_read uses pop() and then expects to write all data to the 2nd buffer.");

        let uart_rx = BufferedUartRx::new(
            uart,
            rx_irq,
            rx_pin,
            rx_buffer,
            uart_config,
        );
        RS485 {
            pio, rx_pin: rx_pin_for_autobaud, tx_pin, tx_en_pin, dma_channel, tx_dma_channel,
            rx: uart_rx,
            rx_buffer: [0; RX_BUF_LEN],
            uart_baud_rate: uart_config.baudrate,
            uart_parity: uart_config.parity,
            handler,
        }
    }

    pub async fn run_task(mut self: Self) {
        let Pio {
            mut common,
            sm0: mut sm,
            mut sm1,
            mut irq0,
            mut irq1,
            ..
        } = Pio::new(self.pio);

        sm.set_enable(false);

        let prg_set_osr = pio_proc::pio_asm!(
            "pull",
            "hang:",
            "jmp hang",
        );

        let prg_autobaud = pio_proc::pio_asm!(
            // Send zero to RX FIFO to signal start of next measurement.
            ".wrap_target",
            "timeout:",
            "in null, 32",

            // Wait for pin to be high for a longer time (UART is idle), copy waiting time to X
            // and decrement until done, start again when pin changes.
            "public wait_idle:",
            "wait 1 pin 0",
            "mov x, osr",
            "continue_wait_idle:",
            "jmp pin still_one",
            "jmp wait_idle",  // we got a zero, start again
            "still_one:",
            "jmp x--, continue_wait_idle",

            // We have a long 1, i.e. UART is idle. Wait for start bit.
            "have_long_high:",
            "irq set 1",          // notify MCU in case they want to know about this (end of Modbus frame)
            "public wait_for_not_idle:",
            ".define public wait_for_not_idle wait_for_not_idle",
            "wait 0 pin 0",
            "nop [4]",
            "jmp pin wait_idle",  // abort if zero was just a glitch

            // We want to measure 9 bits (start bit plus eight (or nine) data bits). However, we need
            // different code for high and low so it's easier to measure 4x2 bits. Weuse Y as the loop counter.
            "set y, 3",

            // Start counting. We count down from the value in OSR (which is our timeout).
            "measure_next:",
            "mov x, osr",
            "measure_low:",
            "jmp pin changed_to_one",
            //"nop",
            "jmp x-- measure_low",
            "jmp timeout",
            "changed_to_one:",
            "in x, 32"

            // Start counting, again, but for a high pulse.
            "mov x, osr",
            "measure_high:",
            "jmp pin still_high2",
            "jmp changed_to_low",
            "still_high2:",
            "jmp x-- measure_high",
            "jmp timeout",
            "changed_to_low:",
            "in x, 32",
            "jmp y-- measure_next",

            // We are done.
            "irq set 0",
            ".wrap",
        );

        // This is based on the UART TX example in the datasheet but we also control the TX_EN pin
        // and we have 9 data bits because we have parity.
        /*
        let prg_tx = pio_proc::pio_asm!(
            ".side_set 2 opt",
            // An 8e1 UART transmit program.
            // OUT pin 0 and side-set pin 1 are both mapped to UART TX pin.
            "pull side 1",           // disable transmitter, get number of bytes
            "mov y, osr side 3",     // copy to y reg, enable transmitter
            "byteloop:",
            "pull side 3 [6]",       // Assert stop bit, or stall with line in idle state (6 instead of 7 because mov/jmp take one clk)
            "set x, 8 side 2 [7]",   // Preload bit counter, assert start bit for 8 clocks
            "bitloop:",              // This loop will run 9 times (8e1 UART, i.e. 8 data bits and one clock bit)
            "out pins, 1",           // Shift 1 bit from OSR to the first OUT pin
            "jmp x-- bitloop [6]",   // Each loop iteration is 8 cycles.
            "jmp y-- byteloop side 3",
        );
        */
        // Let's use 4 cycles for each baud time because we need two bits of side-set.
        let prg_tx = pio_proc::pio_asm!(
            ".origin 24",
            ".side_set 2 opt",
            // An 8e1 UART transmit program.
            // OUT pin 0 and side-set pin 1 are both mapped to UART TX pin.
            "pull side 2",           // disable transmitter, get number of bytes. Side: TX_EN=0, TX=1
            "mov y, osr side 3",     // copy to y reg, enable transmitter
            "byteloop:",
            "pull side 3 [2]",       // Assert stop bit, or stall with line in idle state (6 instead of 7 because mov/jmp take one clk)
            "set x, 8 side 1 [3]",   // Preload bit counter, assert start bit for 8 clocks. Side: TX_EN=1, TX=0
            "bitloop:",              // This loop will run 9 times (8e1 UART, i.e. 8 data bits and one clock bit)
            "out pins, 1",           // Shift 1 bit from OSR to the first OUT pin
            "jmp x-- bitloop [2]",   // Each loop iteration is 8 cycles.
            "jmp y-- byteloop side 3",
        );

        let relocated = RelocatedProgram::new(&prg_set_osr.program);
        let mut cfg = Config::default();
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        const CPU_CLK: u32 = 125_000_000;
        const SM_FREQ: u32 = 25_000_000;
        cfg.clock_divider = U24F8::from_num(CPU_CLK / SM_FREQ);
        cfg.shift_in = ShiftConfig {
            auto_fill: true,
            threshold: 32,
            direction: ShiftDirection::Left,
        };
        cfg.shift_out = ShiftConfig {
            auto_fill: false,
            threshold: 32,
            direction: ShiftDirection::Right,
        };
        let rx_pin_pio = common.make_pio_pin(unsafe { self.rx_pin.clone_unchecked() });
        cfg.set_in_pins(&[&rx_pin_pio]);
        cfg.set_jmp_pin(&rx_pin_pio);

        sm.set_config(&cfg);
        sm.set_enable(true);

        // Timeout: Modbus wants 3.5 char times between frames but we should detect the end of a frame
        // after 1.5 char times. This is 0.86 ms at 19200 baud (8e1, i.e. 11 baud times per char).
        // We need two clocks for each loop.
        //let timeout_start_value = (SM_FREQ as f32 * 0.86e-3 / 2.) as u32;
        let symbols_per_byte = if self.uart_parity == Parity::ParityNone { 10 } else { 11 };
        let timeout_seconds = 1.5 * symbols_per_byte as f32 / self.uart_baud_rate as f32;
        let timeout_start_value = (SM_FREQ as f32 * timeout_seconds / 2.) as u32;
        sm.tx().push(timeout_start_value);
        info!("timeout_start_value: {} = 0x{:08x}, {} sec, clkdiv 0x{:08x}",
            timeout_start_value, timeout_start_value, timeout_seconds, cfg.clock_divider.to_bits());

        // switch to the real program and join FIFOs
        unsafe { common.free_instr(loaded_program.used_memory); };
        sm.set_enable(false);
        let relocated = RelocatedProgram::new(&prg_autobaud.program);
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[]);
        cfg.fifo_join = FifoJoin::RxOnly;
        sm.set_config(&cfg);
        sm.set_enable(true);

        // set rx pin function back to UART
        // (PIO can always read and we don't want to write but embassy_rp cannot know that so it "helpfully" configured the pin for PIO function.)
        unsafe {
            pin_io(&self.rx_pin).ctrl().write(|w| w.set_funcsel(embassy_rp::pac::io::vals::Gpio17ctrlFuncsel::UART0_RX.0));

            // enable pullup to have it not float when we set the transceiver to transmit
            pin_pad(&self.rx_pin).modify(|w| w.set_pue(true));
        }

        // This switching of RX function might have caused an error to be received. Let's skip that.
        if true {
            //NOTE We were using self.rx.fill_buf() but that somehow caused us to not receive any more data afterwards.
            let mut buf = [0; 1];
            let x = select(self.rx.read(&mut buf), Timer::after(Duration::from_micros(10))).await;
            if false {
                match x {
                    Either::First(x) =>
                        info!("DEBUG: first rx: {:?}", x),
                    Either::Second(x) =>
                        info!("DEBUG: first rx: {:?}", x),
                }
            }
            match x {
                Either::First(Ok(_len)) => {
                    //self.rx.consume(len)
                },
                _ => (),
            }
        }

        info!("Program size for auto-baud: {}, for tx: {}, wait_for_not_idle is at: {}",
            prg_autobaud.program.code.len(), prg_tx.program.code.len(),
            prg_autobaud.public_defines.wait_for_not_idle);


        let relocated = RelocatedProgram::new(&prg_tx.program);
        let mut cfg = Config::default();
        let tx_pin_pio = common.make_pio_pin(self.tx_pin);
        let tx_en_pin_pio = common.make_pio_pin(self.tx_en_pin);
        let loaded_program = common.load_program(&relocated);
        cfg.use_program(&loaded_program, &[&tx_en_pin_pio, &tx_pin_pio]);
        cfg.clock_divider = (U56F8::from_num(CPU_CLK) / U56F8::from_num(self.uart_baud_rate*4)).to_fixed();
        cfg.shift_out = ShiftConfig {
            auto_fill: false,
            threshold: 32,
            direction: ShiftDirection::Right,
        };
        cfg.set_out_pins(&[&tx_pin_pio]);
        cfg.fifo_join = FifoJoin::TxOnly;
        sm1.set_config(&cfg);
        sm1.set_pins(gpio::Level::Low, &[&tx_en_pin_pio]);
        sm1.set_pins(gpio::Level::High, &[&tx_pin_pio]);
        sm1.set_pin_dirs(embassy_rp::pio::Direction::Out, &[&tx_en_pin_pio, &tx_pin_pio]);
        sm1.set_enable(true);

        //FIXME Do we have any way to get this number from the outside?
        const TX_BUF_LENGTH: usize = 256;
        defmt::assert!(TX_BUF_LENGTH == H::TX_BUF_LENGTH);
        let mut tx_data = [0u32; TX_BUF_LENGTH+1];
        let mut tx_in_progress = false;

        let mut dma_in_ref = self.dma_channel.into_ref();
        let mut dma_tx_ref = self.tx_dma_channel.into_ref();
        let mut din = [42u32; 9];
        let mut bit_index = 0;
        let mut tx_future = DontAbort::hang();
        loop {
            let x = select4(
                irq0.wait(),
                sm.rx().dma_pull(dma_in_ref.reborrow(), &mut din),
                //debug_print_pio_addr(embassy_rp::pac::PIO0.sm(0)),
                tx_future.continue_wait(),
                select3(
                    //rx_future.continue_wait(),
                    core::future::pending::<()>(),
                    //debug_print_pio_addr(embassy_rp::pac::PIO0.sm(1)),
                    core::future::pending::<()>(),
                    irq1.wait(),
                ),
            ).await;
            match x {
                Either4::First(_) => {
                    let debug_autobaud = false;

                    let mut sum = 0;
                    let mut first = 0;
                    let mut ok = false;
                    let mut got_to_eight = false;
                    for i in 0..din.len() {
                        if din[i] == 0 {
                            bit_index = 0;
                            if debug_autobaud {
                                info!("SM in {}: {:08x} -> start", i, din[i]);
                            }
                        } else {
                            let mut delay = timeout_start_value - din[i];
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            if debug_autobaud {
                                info!("SM in {} ({}): {:08x} -> {} -> {} ms -> {}", i, bit_index, din[i], delay, millis, baud);
                            }
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        }
                    }

                    for _ in bit_index..8 {
                        if let Some(x) = sm.rx().try_pull() {
                            let mut delay = timeout_start_value - x;
                            if bit_index == 0 {
                                delay += 7;
                            } else if bit_index%2 == 1 {
                                //delay += 4;
                                delay += 3;
                            } else {
                                delay += 3;
                            }
                            let millis = (delay) as f32 / SM_FREQ as f32 * 1000. * 2.;
                            let baud = 1000. / millis;
                            if debug_autobaud {
                                info!("SM in ({}): {:08x} -> {} -> {} ms -> {}", bit_index, x, delay, millis, baud);
                            }
                            if bit_index == 0 {
                                sum = 0;
                                first = delay;
                                ok = true;
                            } else if delay < first-first/4 || delay > first+first/4 {
                                ok = false;
                            }
                            sum += delay;
                            bit_index += 1;
                            got_to_eight = bit_index == 8;
                        } else {
                            break
                        }
                    }

                    if got_to_eight && ok {
                        let millis = sum as f32 / SM_FREQ as f32 * 1000. * 2. / 8.;
                        let baud = 1000. / millis;
                        self.handler.on_autobaud_success(baud);
                    }
                },
                Either4::Second(_) => {
                    if false {
                        for i in 0..din.len() {
                            if din[i] == 0 {
                                bit_index = 0
                            } else {
                                bit_index += 1
                            }
                            info!("SM in {}: {:08x}", i, din[i]);
                        }
                    }
                },
                Either4::Third(_) => {
                    tx_in_progress = false;
                    self.handler.on_tx_done();
                },
                Either4::Fourth(Either3::First(_)) => {
                    /*
                    drop(rx_future);
                    let rx_char = x.map(|_| -> u8 { rx_buf_one[0] });
                    if Err(embassy_rp::uart::Error::Overrun) == x {
                        // We should probably drain the FIFO but I don't see a good way to do this.
                    }
                    rx_future = DontAbort::new(self.rx.read(&mut rx_buf_one), PanicIfReused);

                    if tx_in_progress {
                        // use a dummy closure to fix the type of the option because I don't know a better way...
                        let dummy = |_txdata: &[u8]| {};
                        self.handler.on_rx(rx_char, if false {Option::Some(dummy)} else {Option::None});
                    } else {
                        drop(tx_future);
                        tx_future = DontAbort::hang();
                        let mut txlen = Option::None;

                        self.handler.on_rx(rx_char, Option::Some(|txdata: &[u8]| {
                            tx_data[0] = (txdata.len() - 1) as u32;
                            for i in 0..txdata.len() {
                                let x = txdata[i] & 0xff;
                                let mut parity = 0;
                                for j in 0..8 {
                                    parity ^= x>>j;
                                }
                                tx_data[i + 1] = x as u32 | (((parity as u32) & 1) << 8);
                            }
                            txlen = Option::Some(txdata.len());
                        }));
                        if let Option::Some(len) = txlen {
                            //FIXME Wait for idle and then wait again from `wait_idle` before sending. We should have a gap of 3.5 chars.
                            tx_future = DontAbort::new(sm1.tx().dma_push(dma_tx_ref.reborrow(),
                                &tx_data[0..(len+1)]), HangIfReused);
                            tx_in_progress = true;
                        }
                    }
                    */
                },
                Either4::Fourth(Either3::Third(())) => {
                    let mut wait_for_rx_irq_cnt = 0;

                    let mut rx_data = match select(self.rx.read(&mut self.rx_buffer), core::future::ready(())).await {
                        Either::First(Ok(len)) => {
                            // We were sometimes missing the last two chars so wait for the IRQ to copy them to the buffer.
                            while ! unsafe { embassy_rp::pac::UART0.uartfr().read().rxfe() } && wait_for_rx_irq_cnt < 30 {
                                // wait for IRQ to handle this
                                wait_for_rx_irq_cnt += 1;
                                if false {
                                    Timer::after(Duration::from_micros(1)).await;
                                } else {
                                    yield_now().await;
                                }
                            }

                            // The UART uses a ringbuffer and it only pops data once so we need to ask a second
                            // time if the frame wraps around the end of the ringbuffer and we want all data.
                            match select(self.rx.read(&mut self.rx_buffer[len..]), core::future::ready(())).await {
                                Either::First(Ok(len2)) => {
                                    //info!("DEBUG: We needed two calls to self.rx.read(): {} + {}", len, len2);
                                    if false {
                                        // ... and a third time's the charm because we may have received the last char right now.
                                        // -> Actually, it isn't. That's still not enough so we have the loop with rxfe above.
                                        match select(self.rx.read(&mut self.rx_buffer[len+len2..]), core::future::ready(())).await {
                                            Either::First(Ok(len3)) => {
                                                info!("DEBUG: We needed three calls to self.rx.read(): {} + {} + {}", len, len2, len3);
                                                Ok(&self.rx_buffer[0..len+len2+len3])
                                            },
                                            Either::First(Err(err)) => Err(err), // oops, we are ignoring the previous data
                                            Either::Second(()) => Ok(&self.rx_buffer[0..len+len2]),
                                        }
                                    } else {
                                        Ok(&self.rx_buffer[0..len+len2])
                                    }
                                },
                                Either::First(Err(err)) => Err(err), // oops, we are ignoring the previous data
                                Either::Second(()) => Ok(&self.rx_buffer[0..len]),
                            }
                        },
                        Either::First(Err(err)) => Err(err),
                        Either::Second(()) => {
                            Ok(&self.rx_buffer[0..0])
                        },
                    };

                    if wait_for_rx_irq_cnt > 0 {
                        info!("DEBUG: RX: {} bytes, waited for {} cycles", rx_data.map(|x| x.len()), wait_for_rx_irq_cnt);
                    }

                    if let Err(err) = rx_data {
                        // We know that self.rx_buffer isn't borrowed in this arm but rustc doesn't so let's help it a bit.
                        rx_data = Err(err);

                        // We got an error. Let's drain the RX buffers.
                        // How to test: Set baudrate to 115200 and then send data at 19200 baud. Then, test with correct rate.
                        //FIXME This doesn't work quite yet because random errors often cause an overrun error and we need a
                        //      single good char to clear an overrun error. This is a problem because we will keep overrunning
                        //      when we are receiving long frames. It does work as long as no error happens so this is an issue
                        //      with error handling.
                        //      I think the issue may be that on_interrupt() breaks the loop in case of an error but it won't be
                        //      called again unless new data is received. If we do receive new data, the overrun error bit will
                        //      be set for the first byte, the interrupt will stop processing after that and we run into another
                        //      overrun because we will wait for the end of the frame before we handle this.
                        unsafe {
                            //FIXME Use the right UART instead of hardcoding this!
                            while ! embassy_rp::pac::UART0.uartfr().read().rxfe() {
                                let x = embassy_rp::pac::UART0.uartdr().read().0;
                                info!("drain from uartdr: 0x{:08x}", x);
                            }

                            if err == uart::Error::Overrun {
                                //FIXME This doesn't help so we should probably remove it.
                                info!("trying to clear OE bit in UARTRSR");
                                info!("DEBUG: uartrsr=0x{:08x}", embassy_rp::pac::UART0.uartrsr().read().0);
                                embassy_rp::pac::UART0.uartrsr().write(|x| x.set_oe(true));
                                embassy_rp::pac::UART0.uarticr().write(|x| x.set_oeic(true));
                                info!("DEBUG: uartrsr=0x{:08x}", embassy_rp::pac::UART0.uartrsr().read().0);

                                embassy_rp::pac::UART0.uartcr().modify(|w| w.set_rxe(false));
                                embassy_rp::pac::UART0.uartcr().modify(|w| w.set_rxe(true));

                                info!("DEBUG: uartris=0x{:08x}", embassy_rp::pac::UART0.uartris().read().0);
                            }
                        }
                        for _ in 0..10 {
                            match select(self.rx.read(&mut self.rx_buffer), core::future::ready(())).await {
                                Either::First(_) => {
                                    info!("drain");
                                },
                                Either::Second(()) => {
                                    if false {
                                        if ! unsafe { embassy_rp::pac::UART0.uartfr().read().rxfe() } {
                                            info!("drain got nothing but rxfe is not set");
                                            yield_now().await;
                                        } else {
                                            info!("drain got nothing and rxfe is set -> we are done");
                                            break;
                                        }
                                    } else {
                                        info!("drain got nothing -> we are done");
                                        break;
                                    }
                                }
                            }
                        }
                        unsafe {
                            // enable RX interrupts
                            // (We know that there is space in the buffers because we have just cleared them.)
                            //FIXME This doesn't help so we should probably remove it.
                            /*embassy_rp::pac::UART0.uartimsc().write_set(|w| {
                                w.set_rxim(true);
                                w.set_rtim(true);
                            });*/
                            embassy_rp::pac::UART0.uartimsc().modify(|w| {
                                w.set_rxim(true);
                                w.set_rtim(true);
                            });
                        }
                    }

                    //let addr = unsafe { embassy_rp::pac::PIO0.sm(0).addr().read().addr() };
                    //info!("DEBUG: pio {}", addr);

                    if tx_in_progress {
                        // use a dummy closure to fix the type of the option because I don't know a better way...
                        let dummy = |_txdata: &[u8]| {};
                        self.handler.on_rx_frame(rx_data, if false {Option::Some(dummy)} else {Option::None});
                    } else {
                        drop(tx_future);
                        tx_future = DontAbort::hang();
                        let mut txlen = Option::None;

                        self.handler.on_rx_frame(rx_data, Option::Some(|txdata: &[u8]| {
                            tx_data[0] = (txdata.len() - 1) as u32;
                            for i in 0..txdata.len() {
                                let x = txdata[i] & 0xff;
                                let mut parity = 0;
                                for j in 0..8 {
                                    parity ^= x>>j;
                                }
                                tx_data[i + 1] = x as u32 | (((parity as u32) & 1) << 8);
                            }
                            txlen = Option::Some(txdata.len());
                        }));
                        if let Option::Some(len) = txlen {
                            //FIXME Wait for idle and then wait again from `wait_idle` before sending. We should have a gap of 3.5 chars.
                            tx_future = DontAbort::new(sm1.tx().dma_push(dma_tx_ref.reborrow(),
                                &tx_data[0..(len+1)]), HangIfReused);
                            tx_in_progress = true;
                        }
                    }
                },
                _ => {
                },
            }
        }
    }
}
